class SudokuBoard:
    """The board of Sudoku game with utility methods.
    
    The board is basically n^2 x n^2 array of fields.
    
    Board creation and initialization:
    
    - The object is created by
        b = SudokuBoard()
      Then in contains empty board and should be initialized somehow.
      
    - There are 3 methods for board initialization:
        b.from_full_state(state)
          - the state must be list of list (or generaly iterable of iterables),
            each inner list is a row of the board, they should ths have 
            the same length.
        b.from_sparse_state(state)
          - the state must be a dictionary mapping a position to a value.
            E.g. state = {(1,1): 2, (4,1): 3} means that the board should
            contain 2 at position (1,1) and 3 at position (4,1).
    - The method
        b.get_full_state()
      returns the board state in full represenation, i.e. as list of lists.

    Board loading and saving:
    
    - b.load(filepath)
      - loads the board from a file. Each line of board should be 
        on a separate line in the file, having no spaces between numbers.
        Empty fields are marked by '.'.
      - it ignores the lines starting with '-', so that you can use
        e.g. '----------' to separate each triple of lines.
      - it ignores the '|' characters, so that it can be used to separate
        the triples of numbers on the line.
    
    - b.save(filepath)
      - stores the board in the text file formatted as described 
        in the load() method.
    
    Board indexing:
    
    - Use a tuple (x,y), where both x and y are in [1,n], as index 
      to the board, e.g.:
        print(b[(1,1)])
        b[(1,1)] = 2
        
    - The method
        b.free_positions()
      generates the positions of non-assigned fields, one by one.
    
    Accessing fields in certain row, column, or square:
    
    - All the following methods are generators and are thus suitable 
      for use inside for-loops. If you want a list (or another collection),
      enclose the method call inside list() command, e.g.
        list_of_numbers_in_row = list(b.numbers_in_row(row_num))
    
    - The methods
        b.positions_in_row(row_num)
        b.positions_in_column(col_num)
        b.positions_in_square_containing(pos)
      generate a series of positions ((x,y)-tuples) in the particular 
      row, column, or square, respectively.
      
    - Similarly, the methods
        b.numbers_in_row(row_num)
        b.numbers_in_column(col_num)
        b.numbers_in_square_containing(pos)
      generate a series of numbers for nonempty fields in the particular
      row, column, or square, respectively.
      
    - The methods
        b.rows()
        b.columns()
        b.squares()
      generate series of lists of values of the respective groups.
      
    Sudoku-related set of methods:
    - b.candidates(pos)
        returns a set of candidate numbers for the respective position.
    - b.is_correct_solution()
        checks if the board b is a correct solution of Sudoku.
    """
    
    def __init__(self):
        self.size = 0
        self.board = []
                
    def _set_size(self, size):
        self.size = size
        self.base = int(size**0.5)
        
    def _init_board_with_size(self, size):
        self._set_size(size)
        self.board = [[0 for x in range(size)] for x in range(size)]
            
    def from_full_state(self, full):
        """Initialize the board using complete information about the state.
        
        The state must be list of list (or generaly iterable of iterables),
        each inner list is a row of the board, they should ths have 
        the same length.
        """
        self._init_board_with_size(len(full))
        for ir, row in enumerate(full):
            for ic, num in enumerate(row):
                self[(ic+1,ir+1)] = num
            
    def from_sparse_state(self, size, sparse):
        """Initialize the board using sparse representation of the state.
        
        The state must be a dictionary mapping a position to a value.
        E.g. state = {(1,1): 2, (4,1): 3} means that the board should
        contain 2 at position (1,1) and 3 at position (4,1).
        """
        self._init_board_with_size(size)
        for pos, value in sparse.items():
            self[pos] = value
            
    def get_full_state(self):
        """Return the board state in full represenation, i.e. as list of lists."""
        row_tuples = []
        for row in self.board:
            row_tuples.append(tuple(row))
        return tuple(row_tuples)
        
    def clone(self):
        """Return a copy of itself."""
        my_copy = SudokuBoard()
        my_copy.from_full_state(self.get_full_state())
        return my_copy
            
    def load(self, fpath):
        """Load the board from a text file.
        
        Each line of board should be on a separate line in the file, 
        having no spaces between numbers. Empty fields are marked by '.'.
        It ignores the lines starting with '-', so that you can use
        e.g. '----------' to separate each triple of lines.
        It ignores the '|' characters, so that it can be used to separate
        the triples of numbers on the line.
        """
        self.board = []
        with open(fpath, 'rt', encoding='utf-8') as f:
            for line in f:
                if line.startswith('-'): continue
                row = self._process_line(line)
                if row:
                    self.board.append(row)
        self._set_size(len(self.board))
    
    def _process_line(self, line):
        """Helper method, used internally in load()."""
        numbers = []
        for char in line.strip():
            if char == '|': continue
            if char == '.': numbers.append(0)
            else: numbers.append(int(char))
        return numbers
                
    def save(self, fpath):
        """Save the board to a text file."""
        
        with open(fpath, 'wt', encoding='utf-8') as f:
            for i, row in enumerate(self.rows()):
                if i % self.base == 0 and i != 0:
                    f.write('-'*(self.size+self.base-1) + '\n')
                f.write(self._encode_row(row) + '\n')

    def _encode_row(self, row):
        """Helper method, used internally in save()."""
        line = ''
        for i, num in enumerate(row):
            if i % self.base == 0 and i != 0:
                line += '|'
            line += str(num) if num != 0 else '.'
        return line
    
    def __getitem__(self, pos):
        if isinstance(pos, tuple):
            return self.board[pos[1]-1][pos[0]-1]
        else:
            return self.board[pos]
    
    def __setitem__(self, pos, value):
        if isinstance(pos, tuple):
            self.board[pos[1]-1][pos[0]-1] = value
        else:
            self.board[pos] = value
        
    def numbers(self):
        """Generate the numbers for the current board.
        
        Can be used as
        - the numbers that should be filled in each row, column, and square.
        - the indices of columns and rows forming the position (x,y) tuples.
        """
        for i in range(1, self.size+1):
            yield i
            
    def positions_in_row(self, row_number):
        """Generate the positions in a particular row, one by one."""
        for col_number in self.numbers():
            yield (col_number, row_number)
            
    def positions_in_column(self, col_number):
        """Generate the positions in a particular column, one by one."""
        for row_number in self.numbers():
            yield (col_number, row_number)

    def positions_in_square_containing(self, pos):
        """Generate the positions in a particular square, one by one."""
        sx = divmod(pos[0]-1, self.base)[0]
        sy = divmod(pos[1]-1, self.base)[0]
        sx = sx * self.base
        sy = sy * self.base
        for py in range(1, self.base+1):
            for px in range(1, self.base+1):
                yield (sx+px, sy+py)
                
    def numbers_in_row(self, row_number):
        """Generate the numbers already used in a particular row, one by one."""
        for pos in self.positions_in_row(row_number):
            if self[pos] != 0:
                yield self[pos]
            
    def numbers_in_column(self, col_number):
        """Generate the numbers already used in a particular column, one by one."""
        for pos in self.positions_in_column(col_number):
            if self[pos] != 0:
                yield self[pos]
            
    def numbers_in_square_containing(self, cur_pos):
        """Generate the numbers already used in a particular square, one by one."""
        for pos in self.positions_in_square_containing(cur_pos):
            if self[pos] != 0:
                yield self[pos]

    def rows(self):
        """Generate a list of the values of fields in a row, row by row."""
        for row in self.board:
            yield row
            
    def columns(self):
        """Generate a list of the values of fields in a column, column by column."""
        for icol in range(self.size):
            yield [row[icol] for row in self.board]
            
    def squares(self):
        """Generate a list of the values of fields in a square, square by square."""
        for iy in range(self.base):
            rows = self.board[iy*self.base:(iy+1)*self.base]
            for ix in range(self.base):
                group = []
                for row in rows:
                    group.extend(row[ix*self.base:(ix+1)*self.base])
                yield group

    def __str__(self):
        value = ''
        for row in self.rows():
            value += str(row) + '\n'
        return value
    
    def free_positions(self):
        """Generate the positions of free fields, one by one."""
        for icol in self.numbers():
            for irow in self.numbers():
                pos = (icol,irow)
                if self[pos] == 0:
                    yield pos
                    
    def fixed_positions(self):
        """Generate the positions of already fixed fields, one by one."""
        for icol in self.numbers():
            for irow in self.numbers():
                pos = (icol,irow)
                if self[pos] != 0:
                    yield pos
                    
    def is_filled(self):
        """Check if a number is assigned to all positions."""
        if len(list(self.free_positions()))==0:
            return True
        return False
                    
    def candidates(self, pos):
        """Return a set of candidate numbers for a particulat position."""
        if self[pos] != 0:
            return set()
        numbers_left = set(self.numbers())
        numbers_left -= set(self.numbers_in_column(pos[0]))
        numbers_left -= set(self.numbers_in_row(pos[1]))
        numbers_left -= set(self.numbers_in_square_containing(pos))
        return numbers_left
        
    def is_correct_solution(self):
        """Check if the current board state represents a correct solution."""
        for row in self.rows():
            if not self._group_contains_all_numbers(row):
                return False
        for col in self.columns():
            if not self._group_contains_all_numbers(col):
                return False
        for square in self.squares():
            if not self._group_contains_all_numbers(square):
                return False
        return True

    def _group_contains_all_numbers(self, group):
        numbers_left = set(self.numbers())
        numbers_left -= set(group)
        if not numbers_left:
            return True
        else:
            return False

    def is_ancestor_of(self, other):
        """Check if the 'other' SudokuBoard can be a descendant of the current board."""
        for pos in self.fixed_positions():
            if self[pos] != other[pos]:
                return False
        return True
    
    def is_correct_solution_of(self, task_board):
        """Checks if the current board is a correct solution to a given task board."""
        if not task_board.is_ancestor_of(self):
            return False
        return self.is_correct_solution()
